import os
import yatr.base as ybase

from jinja2 import UndefinedError
from nose.tools import assert_raises
from syn.base_utils import assign, capture, setitem
from yatr.base import resolve, variables, ordered_macros, get_output,\
    str_to_bool, fix_functions, args_kwargs_from_env, expand_path

#-------------------------------------------------------------------------------
# Utilities

def test_expand_path():
    assert expand_path('/foo/bar') == '/foo/bar'

    home = os.path.expanduser('~')
    assert expand_path('~/foo/bar') == '{}/foo/bar'.format(home)

    with setitem(os.environ, 'FOO_BAR', 'baz'):
        assert expand_path('~/foo/$FOO_BAR') == '{}/foo/baz'.format(home)
        assert expand_path('~/foo/${FOO_BAR}_fix.yml') == \
            '{}/foo/baz_fix.yml'.format(home)

def test_resolve():
    assert resolve('abc', {}) == 'abc'
    assert_raises(UndefinedError, resolve, 'ab{{c}}', {})
    assert resolve('ab{{c}}', {}, lenient=True) == 'ab'
    assert resolve('ab{{c}}', dict(c='d')) == 'abd'

    # No autoescape
    assert resolve('"abc" <in >& /dev/null', {}) == '"abc" <in >& /dev/null'

    # list macros
    assert resolve(['{{a}}', '{{b}}', 'c'], dict(a='d', b='e')) == ['d', 'e', 'c']

    # dict macros
    assert resolve(dict(a='{{b}}', b='1'), dict(b='2')) == dict(a='2', b='1')
    assert resolve('{{a.b}}', dict(a=dict(b='c'))) == 'c'

def test_variables():
    assert variables('{{a}} {{b}}') == {'a', 'b'}
    assert variables(['{{a}}', '{{b}}', 'c']) == {'a', 'b'}
    assert_raises(Exception, variables, '{{a|foo}}')

def test_ordered_macros():
    assert list(ordered_macros({})) == []

    dct = dict(a = 'b')
    assert list(ordered_macros(dct)) == [('a', 'b')]

    dct = dict(a = 'b',
               b = '{{a}}',
               c = '{{b}}')
    assert list(ordered_macros(dct)) == [('a', 'b'), 
                                         ('b', '{{a}}'),
                                         ('c', '{{b}}')]

    dct = dict(a = 'b',
               b = '{{a}}',
               c = ['{{b}}', 'c'],
               d = '{{c}}')
    assert list(ordered_macros(dct)) == [('a', 'b'), 
                                         ('b', '{{a}}'),
                                         ('c', ['{{b}}', 'c']),
                                         ('d', '{{c}}')]

def test_get_output():
    out, code = get_output('true')
    assert out == ''
    assert code == 0

    out, code = get_output('false')
    assert out == ''
    assert code == 1

    with capture() as (out_, err):
        out, code = get_output('python -c "raise Exception(\\"Test\\")"')
        assert 'Exception' in out
        assert code == 1

def test_str_to_bool():
    assert str_to_bool(1) is True
    assert str_to_bool(True) is True
    assert str_to_bool(0) is False
    assert str_to_bool(False) is False

    assert str_to_bool(' YES   ')
    assert str_to_bool(' TrUE   ')
    assert str_to_bool(' 1   ')

    assert not str_to_bool(' NO   ')
    assert not str_to_bool(' FalSE   ')
    assert not str_to_bool(' 0   ')

    assert_raises(TypeError, str_to_bool, [])
    assert_raises(TypeError, str_to_bool, 'foo')

def test_fix_functions():
    from yatr import Env

    class Foo(object):
        pass
    
    def fooer():
        ret = Foo()
        ret.hex = 'foo'
        return ret

    def identity(s=None, **kwargs):
        return s

    env = Env(macros=dict(abc='def'),
              jinja_functions=dict(abc=identity))
    with capture() as (out, err):
        with assign(ybase, 'uuid4', fooer):
            out = fix_functions("{{abc('sdf')}} {{abc}} {{abc('qwe')}}", 
                                {'abc'}, env)
            assert out == "{{abc_foo('sdf')}} {{abc}} {{abc_foo('qwe')}}"
            assert resolve(out, env.macros, jenv=env.jenv) == 'sdf def qwe'

            out = fix_functions("{{abc('ghi')}}", {'abc'}, env)
            assert out == "{{abc_foo('ghi')}}"
            assert resolve(out, env.macros, jenv=env.jenv) == 'ghi'

def test_args_kwargs_from_env():
    assert list(args_kwargs_from_env({})) == [(), {}]
    assert list(args_kwargs_from_env(dict(_1='a',
                                          _2='b',
                                          _3='c',
                                          a='aa',
                                          b='bb'))) == \
        [('a', 'b', 'c'), dict(a='aa', b='bb')]

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



import os
import re
import sys
import shutil
import hashlib
import requests
from uuid import uuid4
from tempfile import mkstemp, mkdtemp
from contextlib import contextmanager
from subprocess import call, check_output, CalledProcessError, STDOUT
from jinja2 import Template, Environment, meta, StrictUndefined, Undefined
from syn.base_utils import Precedes, topological_sorting, assign
from syn.five import STR

DEFAULT_CACHE_DIR = '~/.yatr'
ARG_MACRO_PATTERN = re.compile('^_[0-9]+$')

#-------------------------------------------------------------------------------

class ValidationError(Exception):
    pass

#-------------------------------------------------------------------------------
# Jinja Filters

DEFAULT_JINJA_FILTERS = dict()

#-------------------------------------------------------------------------------
# Jinja Functions

def jfunc_commands(name, **kwargs):
    env = kwargs.pop('env')
    task = env.tasks[name]
    ret = task.run_preview(env, **kwargs)
    return ret

def jfunc_env(name, value=None, **kwargs):
    if value is not None:
        return os.environ.get(name, value)
    return os.environ[name]

DEFAULT_JINJA_FUNCTIONS = dict(commands = jfunc_commands,
                               env = jfunc_env)

#-------------------------------------------------------------------------------
# Utilities

def eprint(out, flush=True):
    sys.stderr.write(out + '\n')
    if flush:
        sys.stderr.flush()

def expand_path(path):
    return os.path.expanduser(os.path.expandvars(path))

def resolve(template, env, lenient=False, jenv=None):
    if isinstance(template, dict):
        return {key: resolve(item, env, lenient, jenv)
                for key, item in template.items()}

    if isinstance(template, list):
        return [resolve(elem, env, lenient, jenv) for elem in template]

    if isinstance(template, STR):
        if '{{' in template:
            if lenient:
                if jenv:
                    with assign(jenv, 'undefined', Undefined):
                        return jenv.from_string(template).render(env)
                return Template(template).render(env)
            else:
                if jenv is None:
                    jenv = Environment(undefined=StrictUndefined)
                return jenv.from_string(template).render(env)
    return template

def variables(template, lenient=False, jenv=None):
    if isinstance(template, dict):
        ret = set()
        for key, item in template.items():
            ret.update(variables(item, lenient, jenv))
        return ret

    if isinstance(template, list):
        ret = set()
        for elem in template:
            ret.update(variables(elem, lenient, jenv))
        return ret

    if jenv is None:
        jenv = Environment()

    try:
        return meta.find_undeclared_variables(jenv.parse(template))
    except Exception as e:
        if lenient:
            return set()
        raise e

def order_relations_from_macros(macros, lenient=False, jenv=None):
    out = []
    for name, template in macros.items():
        for var in variables(template, lenient=lenient, jenv=jenv):
            out.append(Precedes(var, name))
    return out

def ordered_macros(macros, lenient=False, funcs=None, jenv=None):
    funcs = funcs if funcs else []
    rels = order_relations_from_macros(macros, lenient=lenient, jenv=jenv)
    names = topological_sorting(macros, rels)

    for name in names:
        if name in macros:
            yield name, macros[name]
        elif name in funcs:
            pass
        else:
            if not lenient:
                if not re.match(ARG_MACRO_PATTERN, name):
                    raise ValidationError('Referenced macro {} not defined'
                                          .format(name))

def get_output(cmd):
    try:
        code = 0
        out = check_output(cmd, shell=True, stderr=STDOUT)
    except CalledProcessError as e:
        code = e.returncode
        out = e.output
        if out:
            print(out)

    out = out.decode('utf-8') if out else ''
    return out, code

def command(cmd, silent=False):
    if silent:
        out, code = get_output(cmd)
    else:
        code = call(cmd, shell=True)
    return code

def download(url, path):
    req = requests.get(url)
    with open(path, 'wb') as f:
        for data in req.iter_content():
            f.write(data)

def cached_path(path, cachedir=DEFAULT_CACHE_DIR):
    fname = hashlib.sha224(path.encode('utf-8')).hexdigest()
    return os.path.join(cachedir, fname)

def resolve_url(url, cachedir=DEFAULT_CACHE_DIR, force=False):
    if '://' in url:
        cachedir = expand_path(cachedir)

        if not os.path.isdir(cachedir):
            os.mkdir(cachedir)

        fpath = cached_path(url, cachedir)

        if not os.path.isfile(fpath) or force:
            download(url, fpath)

        return fpath
    return url

def read(path):
    with open(path, 'r') as f:
        return f.read()

@contextmanager
def tempfile(*args, **kwargs):
    try:
        _, path = mkstemp(*args, **kwargs)
        yield path
    finally:
        os.remove(path)

@contextmanager
def tempdir(*args, **kwargs):
    try:
        path = mkdtemp(*args, **kwargs)
        yield path
    finally:
        shutil.rmtree(path)

def str_to_bool(value):
    if isinstance(value, int):
        return bool(value)
    elif isinstance(value, STR):
        val = value.strip().lower()
        if val in {'yes', 'true', '1'}:
            return True
        if val in {'no', 'false', '0'}:
            return False
    raise TypeError('Invalid type/value for conversion: {}'.format(value))

def get_delete(in_, out, key, default, outkey=None):
    if outkey is None:
        outkey = key
    out[outkey] = in_.pop(key, default)

def fix_functions(template, potential_problems, env):
    if isinstance(template, list):
        return [fix_functions(elem, potential_problems, env)
                for elem in template]

    fixed = template
    for var in potential_problems:
        test = '(\{\{[^}]*\\b)' + var + '(\([^}]*\)[^}]*\}\})'
        if re.search(test, fixed):
            if var in env.function_aliases:
                alias = env.function_aliases[var]
            else:
                # TODO: replace with real gensym
                alias = var + '_' + uuid4().hex
                env.jenv.globals[alias] = env.jenv.globals[var]
                env.function_aliases[var] = alias
                eprint("Warning: '{}' defined as both macro and Jinja2 "
                       "function.  Using function alias '{}'".format(var, alias))
                
            repl = '\\1' + alias + '\\2'
            fixed = re.sub(test, repl, fixed)

    return fixed

def args_kwargs_from_env(env):
    arg_macros = [name for name in env if re.match(ARG_MACRO_PATTERN, name)]
    args = tuple([env[name] for name in sorted(arg_macros)])
    kwargs = {name: env[name] for name in env if name not in arg_macros}
    return args, kwargs

#-------------------------------------------------------------------------------
# __all__

__all__ = ('ValidationError',)

#-------------------------------------------------------------------------------



import re
import sys
import shlex
from itertools import product
from six.moves import cStringIO
from uuid import uuid4

from syn.base import Base, Attr
from syn.five import STR
from syn.base_utils import assign, message
from syn.type import Callable, Dict, List, This

from .base import ValidationError, get_delete, eprint
from .base import command as command_

#-------------------------------------------------------------------------------

FOREX = re.compile('([a-zA-Z_][a-zA-Z0-9_-]*)\s+in\s+([a-zA-Z_][a-zA-Z0-9_-]*)')

#-------------------------------------------------------------------------------
# For


class For(Base):
    _attrs = dict(var = Attr((STR, List(STR)), doc='The loop variable(s)'),
                  in_ = Attr((STR, List((STR, int, list))), 
                             doc='Name(s) of list macro(s) to loop over'))
    _opts = dict(init_validate = True, 
                 args = ('var', 'in_'))

    @classmethod
    def from_yaml(cls, dct):
        if isinstance(dct, STR):
            m = re.match(FOREX, dct)
            if m:
                return cls(var=m.groups()[0], in_=m.groups()[1])
            else:
                raise ValidationError('Invalid for loop specifier: {}'
                                      .format(dct))

        kwargs = {}
        get_delete(dct, kwargs, 'var', None)
        get_delete(dct, kwargs, 'in', None, 'in_')
        
        if dct:
            raise ValidationError('Invalid for keys: {}'
                                  .format(','.join(dct.keys())))
        return cls(**kwargs)

    def resolve_macros(self, env, **kwargs):
        var = self.var
        in_ = self.in_
        if not isinstance(var, list):
            var = [self.var]
            in_ = [self.in_]

        outs = []
        for k, v in enumerate(var):
            name = in_[k]
            if isinstance(name, list):
                val = name
            else:
                val = env.env[name]
                if not isinstance(val, list):
                    raise ValidationError('For loop "in" specifier must be name of '
                                          'list macro: {}'.format(in_[k]))
            outs.append(val)
        return var, outs

    def loop(self, env, **kwargs):
        var, in_ = self.resolve_macros(env, **kwargs)
        
        for tup in product(*in_):
            yld = env.copy(**kwargs)
            for name, val in zip(var, tup):
                yld.env[name] = val
            yield yld

    def validate(self):
        super(For, self).validate()
        
        if isinstance(self.var, list):
            if len(self.var) != len(self.in_):
                raise ValidationError('"var" and "in" lists must be same '
                                      'length')


#-------------------------------------------------------------------------------
# Command


class Command(Base):
    _attrs = dict(command = Attr((STR, Callable)),
                  context = Attr(STR, '')) # string used to look up context object in env
    _opts = dict(init_validate = True,
                 args = ('command',))

    def resolve_macros(self, env, **kwargs):
        command = env.resolve(self.command, **kwargs)
        context = env.resolve(kwargs.get('context', self.context))
        return command, context

    def run(self, env, **kwargs):
        verbose = kwargs.get('verbose', False)
        preview = kwargs.get('preview', False)
        # Called from Task.run_preview()
        run_preview = kwargs.get('run_preview', False)
        silent = kwargs.get('silent', env.settings.get('silent', False))

        pre = ''
        if preview and not run_preview:
            pre = kwargs.get('preview_pre', '')

        command, context_name = self.resolve_macros(env, **kwargs)
        if not context_name:
            context = env.default_context
        else:
            context = env.contexts[context_name]
            
        if callable(context.run):
            return context.run(command, env, **kwargs)

        cmd = context.run_command(command, env, **kwargs)

        if verbose:
            sys.stdout.write(pre + cmd + '\n')
            sys.stdout.flush()
        
        if not preview:
            args = shlex.split(cmd)
            if args[0] == 'yatr':
                try:
                    from .main import _main
                    _main(*args[1:])
                    return 0
                except Exception as e:
                    eprint(message(e))
                    return 1
            return command_(cmd, silent)


#-------------------------------------------------------------------------------
# Task


class Task(Base):
    _attrs = dict(commands = Attr(List(Command)),
                  condition = Attr(This, optional=True),
                  loop = Attr(For, optional=True),
                  args = Attr(List((STR, int)), init=lambda self: list()),
                  kwargs = Attr(Dict((STR, int)), init=lambda self: dict()),
                  condition_type = Attr(bool, True))
    _opts = dict(init_validate = True,
                 optional_none = True)

    @classmethod
    def from_yaml(cls, name, dct):
        if isinstance(dct, STR):
            return cls(commands=[Command(dct)])

        elif List((STR, dict)).query(dct):
            cmds = []
            for s in dct:
                if isinstance(s, STR):
                    cmds.append(Command(s))

                # s is dict at this point
                elif list(s.keys()) == ['task']:
                    # TODO: replace with real gensym
                    name_ = uuid4().hex
                    cmds.append(Task.from_yaml(name_, s['task']))

                else:
                    raise ValidationError('Bad task definition element: {}'
                                          .format(s))
            return cls(commands=cmds)

        elif isinstance(dct, dict):
            if set(dct.keys()).issuperset({'command'}):
                ret = Task.from_yaml(name, dct['command'])

                if 'context' in dct:
                    for cmd in ret.commands:
                        cmd.context = dct['context']

                if 'args' in dct:
                    ret.args = dct['args']

                if 'kwargs' in dct:
                    ret.kwargs = dct['kwargs']

                if 'for' in dct:
                    ret.loop = For.from_yaml(dct['for'])

                if 'if' in dct:
                    ret.condition = Task.from_yaml(name + '-if', dct['if'])
                    ret.condition_type = True
                elif 'ifnot' in dct:
                    ret.condition = Task.from_yaml(name + '-ifnot', dct['ifnot'])
                    ret.condition_type = False

                ret.validate()
                return ret

        raise ValidationError('Invalid data for task: {}'.format(name))

    def run(self, env, **kwargs):
        codes = []
        looping = kwargs.get('looping', False)
        exit_on_error = kwargs.get('exit_on_error',
            env.settings.get('exit_on_error', True))
        preview_conditionals = kwargs.get('preview_conditionals',
            env.settings.get('preview_conditionals', True))

        if self.condition and not looping:
            if preview_conditionals:
                kwargs['preview_pre'] = 'if: ' if self.condition_type else 'ifnot: '
            codes_ = self.condition.run(env, **kwargs)
            code = max(codes_)
            if (((self.condition_type is True and code != 0) or
                 (self.condition_type is False and code == 0)) and 
                code is not None):
                return []
            if preview_conditionals:
                kwargs['preview_pre'] = '\t'

        if (self.args or self.kwargs) and not looping:
            env = env.copy(**kwargs)

            if self.args:
                for k, arg in enumerate(self.args):
                    env.env['_{}'.format(k + 1)] = arg

            if self.kwargs:
                for name, value in self.kwargs.items():
                    env.env[name] = value

        if self.loop and not looping:
            n = 0
            kwargs['looping'] = True
            for env_ in self.loop.loop(env, **kwargs):
                env_.env[env_.settings['loop_count_macro']] = n
                codes_ = self.run(env_, **kwargs)
                codes.extend(codes_)
                n += 1
            return codes

        for cmd in self.commands:
            if isinstance(cmd, Task):
                codes_ = cmd.run(env, **kwargs)
                codes.extend(codes_)

            # cmd is type Command at this point
            elif cmd.command in env.tasks:
                codes_ = env.tasks[cmd.command].run(env, **kwargs)
                codes.extend(codes_)

            else:
                code = cmd.run(env, **kwargs)
                codes.append(code)

            if exit_on_error and any(c != 0 for c in codes if c is not None):
                break

        return codes

    def run_preview(self, env, **kwargs):
        kwargs['preview'] = True
        kwargs['verbose'] = True
        kwargs['run_preview'] = True
        
        with assign(sys, 'stdout', cStringIO()):
            self.run(env, **kwargs)
            ret = sys.stdout.getvalue()
        return ret


Task._attrs.types['commands'] = List((Command, Task))

#-------------------------------------------------------------------------------
# __all__

__all__ = ('For', 'Command', 'Task')

#-------------------------------------------------------------------------------



import os
import sys
from yatr import Command, Task, Env
from yatr.base import read, tempfile

#-------------------------------------------------------------------------------
# Task

def test_task():
    env = Env()
    with tempfile() as f:
        c = Command('python --version > ' + f + ' 2>&1')
        t = Task(name='foo', commands=[c])
        
        codes = t.run(env)
        strs = read(f)

        assert codes == [0]
        assert strs.split()[0] == 'Python'
        assert strs.split()[1] == sys.version.split()[0]
        
        # Test conditional execution
        t = Task.from_yaml('foo', {'command': 'pwd 2>&1 > ' + f,
                                   'if': 'true'})
        codes = t.run(env)
        assert read(f) == os.getcwd() + '\n'
        assert codes == [0]

        t = Task.from_yaml('foo', {'command': 'pwd 2>&1 > ' + f,
                                   'if': 'false'})
        codes = t.run(env)
        assert codes == []

        t = Task.from_yaml('foo', {'command': 'pwd 2>&1 > ' + f,
                                   'ifnot': 'true'})
        codes = t.run(env)
        assert codes == []

        t = Task.from_yaml('foo', {'command': 'pwd 2>&1 > ' + f,
                                   'ifnot': 'false'})
        codes = t.run(env)
        assert read(f) == os.getcwd() + '\n'
        assert codes == [0]

        # Test exit_on_error
        t = Task.from_yaml('foo', ['true', 'false', 'true'])
        codes = t.run(env)
        assert codes == [0, 1]

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



from functools import wraps
from syn.base import Attr, Base
from syn.base_utils import message
from syn.five import STR
from .base import eprint
from .env import Env
from .task import Command
from .task import Task as Task_

#-------------------------------------------------------------------------------
# EnvDecorator


class EnvDecorator(Base):
    _attrs = dict(name = Attr(STR),
                  parent = Attr(Env))
    _opts = dict(init_validate = True,
                 args = ('parent', 'name'))

    def __call__(self, func):
        raise NotImplementedError


#-------------------------------------------------------------------------------
# JinjaFilter


class JinjaFilter(EnvDecorator):
    def __call__(self, func):
        self.parent.jinja_filters[self.name] = func
        return func


#-------------------------------------------------------------------------------
# JinjaFunction


class JinjaFunction(EnvDecorator):
    def __call__(self, func):
        self.parent.jinja_functions[self.name] = func
        return func


#-------------------------------------------------------------------------------
# Task


class Task(EnvDecorator):
    _attrs = dict(display = Attr(tuple, (), 
                                 'Names of kwargs to display in verbose'))

    def __call__(self, f):
        @wraps(f)
        def func(env, *args, **kwargs):
            try:
                f(env, *args, **kwargs)
                return 0

            except Exception as e:
                eprint(message(e))
                return 1

        func.display = self.display
        task = Task_(commands=[Command(func, context='python_callable')],
                     name=self.name)
        self.parent.tasks[self.name] = task
        return func


#-------------------------------------------------------------------------------
# __all__

__all__ = ()

#-------------------------------------------------------------------------------



from copy import deepcopy
from nose.tools import assert_raises

from yatr import Env
from yatr import env as ye
from yatr.env import Updateable, DefaultList
from syn.base_utils import assign, assert_equivalent

#-------------------------------------------------------------------------------
# DefaultList

def test_defaultlist():
    l = DefaultList()
    assert l[0] == 0
    assert l[1] == 0
    
    l.append(1)
    assert l[0] == 1
    assert l[1] == 0

    del l[0]
    del l[0]
    assert l[0] == 0
    assert l[1] == 0

    l = DefaultList([1, 2, 3], default=42)
    assert l[0] == 1
    assert l[1] == 2
    assert l[2] == 3
    assert l[3] == 42

#-------------------------------------------------------------------------------
# Env

def test_env():
    Updateable()._update_pre({})
    Updateable()._update_post({})

    with assign(ye, 'INITIAL_MACROS', dict(a='1', b='2')):
        e = Env()
        
        assert e.macros
        assert e.contexts
        assert e.default_context
        assert not e.tasks
        assert not e.secret_values

    e1 = Env(macros = dict(a='b',
                           b='{{a}}c',
                           c='d{{b}}'))
    e2 = Env(macros = dict(a='a',
                           d='{{c}}e'))
    e3 = Env(macros = dict(a='{{d}}'))

    e1c = e1.copy()
    assert_equivalent(e1, e1c)
    assert_equivalent(e1.macros, e1c.macros)

    e0 = deepcopy(e1)
    e0.resolve_macros()
    assert e0.env == dict(a='b',
                          b='bc',
                          c='dbc')

    e1.update(e2)
    e1c = deepcopy(e1)
    e1.resolve_macros()
    assert e1.env == dict(a='a',
                          b='ac',
                          c='dac',
                          d='dace')

    e2.macros['a'] = 'c'
    e1c.update(e2)
    e1c2 = deepcopy(e1c)
    e1c.resolve_macros()
    assert e1c.env == dict(a='c',
                           b='cc',
                           c='dcc',
                           d='dcce')

    e1c2.update(e3)
    assert_raises(ValueError, e1c2.resolve_macros)

    e = Env(default_task='foo')
    e.update(Env())
    assert e.default_task == 'foo'
    e.update(Env(default_task='bar'))
    assert e.default_task == 'bar'

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



import os
import sys
import shutil
from jinja2 import UndefinedError
from mock import MagicMock
from nose.tools import assert_raises
from syn.base_utils import capture, assign, chdir, setitem
from yatr.main import _main, main, search_rootward, find_bash_completions,\
    dump_bash_completions, BASH_COMPLETION_MESSAGE, USAGE, OPTION_STRINGS,\
    DEFAULT_SETTINGS
from yatr import __version__ as yver
from yatr import base as ybase
from yatr.base import cached_path, read, tempdir, ValidationError, tempfile

DIR = os.path.abspath(os.path.dirname(__file__))
TEST1 = os.path.join(DIR, 'test1.yml')
TEST2 = os.path.join(DIR, 'test2.yml')
TEST3 = os.path.join(DIR, 'test3.yml')
TEST4 = os.path.join(DIR, 'test4.yml')
TEST5 = os.path.join(DIR, 'test5.yml')
TEST6 = os.path.join(DIR, 'test6.yml')
TEST7 = os.path.join(DIR, 'test7.yml')
TEST8 = os.path.join(DIR, 'test8.yml')
TEST8_2 = os.path.join(DIR, 'test8_2.yml')
TEST8_IN = os.path.join(DIR, 'test8.j2')
TEST8_OUT = os.path.join(DIR, 'test8.bash')
TEST9 = os.path.join(DIR, 'test9.yml')
TEST10 = os.path.join(DIR, 'test10.yml')
TEST11 = os.path.join(DIR, 'test11.yml')
TEST12 = os.path.join(DIR, 'test12.yml')
TEST13 = os.path.join(DIR, 'test13.yml')
DOCKERFILE = os.path.join(DIR, 'example/render/Dockerfile')
OUT = os.path.join(DIR, 'output')
URL = 'https://raw.githubusercontent.com/mbodenhamer/yatrfiles/master/yatrfiles/test/test1.yml'
HOME = os.path.expanduser('~')
TEST1_HOME_J2 = os.path.join(HOME, 'BAZ_test1.j2')
TEST1_HOME_TXT = os.path.join(HOME, 'BAZ_test1.txt')
TEST2_HOME = os.path.join(HOME, 'BAZ_test2.yml')
CACHEDIR_HOME = os.path.join(HOME, 'BAZ_cachedir')

def listify(s):
    return [o.strip() for o in s.strip().split()]

#-------------------------------------------------------------------------------

DFRENDER = '''# -*- dockerfile -*-
FROM python:2-alpine

RUN pip install -U --no-cache \\
    syn>=0.0.14

CMD ["python2"]'''

#-------------------------------------------------------------------------------
# Bash completions

def test_bash_completions():
    settings = DEFAULT_SETTINGS
    
    with chdir(DIR):
        out = find_bash_completions(['--cache-dir', '/app', 'print'], 1)
        assert out == []

        out = find_bash_completions(['-f', '/app', 'print'], 1)
        assert out == []

        out = find_bash_completions(['--yatrfile', '/app', 'print'], 1)
        assert out == []


        out = find_bash_completions(['-'], 0)
        assert out == OPTION_STRINGS

        out = find_bash_completions(['p'], 0)
        assert out == ['print']

        out = find_bash_completions(['xyz'], 0)
        assert out == []

        out = find_bash_completions(['-m', 'a'], 1)
        assert out == ['a']

        out = find_bash_completions(['--macro', 'a'], 1)
        assert out == ['a']

        out = find_bash_completions(['-m', 'a='], 1)
        assert out == []

        out = find_bash_completions(['-s', 's'], 1)
        assert out == ['silent']

        out = find_bash_completions(['--setting', 's'], 1)
        assert out == ['silent']

        out = find_bash_completions(['-s', 'silent='], 1)
        assert out == []

        out = find_bash_completions(['--dump-path', 'p'], 1)
        assert out == ['print']

        out = find_bash_completions(['print', 'xyz'], 1)
        assert out == []
        
        out = find_bash_completions(['-i', '/app'], 1)
        assert out == []

        out = find_bash_completions(['-o', '/app'], 1)
        assert out == []


        out = find_bash_completions([], 0)
        assert out == ['print']

        out = find_bash_completions(['-m'], 1)
        assert out == ['a', 'b', 'c']

        out = find_bash_completions(['--macro'], 1)
        assert out == ['a', 'b', 'c']

        out = find_bash_completions(['-s'], 1)
        assert out == settings

        out = find_bash_completions(['--setting'], 1)
        assert out == settings

        out = find_bash_completions(['-m', 'a=xyz'], 2)
        assert out == ['print']

        out = find_bash_completions(['-m', 'a=xyz', 'print'], 3)
        assert out == []

        out = find_bash_completions(['-i'], 1)
        assert out == []

        out = find_bash_completions(['-o'], 1)
        assert out == []

        with capture() as (out, err):
            dump_bash_completions(['-m'], 1)
        assert out.getvalue() == 'a b c\n'

        with capture() as (out, err):
            _main('--dump-bash-completions', '1', '--macro')
        assert out.getvalue() == 'a b c\n'

    with tempdir() as d:
        with chdir(d):
            out = find_bash_completions(['-'], 0)
            assert out == OPTION_STRINGS

            out = find_bash_completions([], 0)
            assert out == []

            out = find_bash_completions(['-m'], 1)
            assert out == []

            out = find_bash_completions(['-s'], 1)
            assert out == settings

            yfpath = os.path.join(DIR, 'yatrfile.yml')
            out = find_bash_completions(['-f', yfpath], 2)
            assert out == ['print']

            out = find_bash_completions(['--yatrfile', yfpath], 2)
            assert out == ['print']

            out = find_bash_completions(['-f', yfpath, '-s'], 3)
            assert out == settings

            yfpath = os.path.join(DIR, 'yatrfile.yaml')
            out = find_bash_completions(['-f', yfpath], 2)
            assert out == []
            
            yfpath = os.path.join(d, 'yatrfile.yml')
            shutil.copyfile(os.path.join(DIR, 'test2.yml'), yfpath)
            out = find_bash_completions([], 0)
            assert out == ['bad', 'good']
            
            with tempdir() as d2:
                cd = os.path.join(d2, 'foo')
                out = find_bash_completions(['--cache-dir', cd], 2)
                assert out == ['bad', 'good']
                
                with open(yfpath, 'r') as f:
                    data = f.read()
                out = data.replace('bad', 'dad')
                with open(yfpath, 'w') as f:
                    f.write(out)

                out = find_bash_completions(['--cache-dir', cd], 2)
                assert out == ['dad', 'good']

    spath = os.path.join(os.path.abspath(os.path.dirname(ybase.__file__)),
                         'scripts/completion/yatr')
    dpath = '/etc/bash_completion.d/yatr'
    with assign(shutil, 'copyfile', MagicMock()):
        with capture() as (out, err):
            _main('--install-bash-completions')
        assert out.getvalue() == BASH_COMPLETION_MESSAGE + '\n'
        assert shutil.copyfile.call_count == 1
        shutil.copyfile.assert_any_call(spath, dpath)

#-------------------------------------------------------------------------------
# main

def test_main():
    assert_raises(RuntimeError, search_rootward, '/')

    with assign(sys, 'exit', lambda *args, **kwargs: None):
        with capture() as (out, err):
            _main('--version')
        assert out.getvalue() == 'yatr {}\n'.format(yver)

        with assign(sys, 'argv', ['', '--version']):
            with capture() as (out, err):
                main()
        assert out.getvalue() == 'yatr {}\n'.format(yver)

        with capture() as (out, err):
            _main('-f', os.path.join(DIR, 'yatrfile1.yml'), '--validate')
        assert out.getvalue() == 'Validation successful\n'

        # Test default task
        with capture() as (out, err):
            _main('-f', TEST2, '-p')
        assert out.getvalue() == 'true\n'

        # Test -f path expansion
        assert not os.path.exists(TEST2_HOME), \
            "Run these tests in a Docker container to avoid file clobbering"

        shutil.copyfile(TEST2, TEST2_HOME)
        with setitem(os.environ, 'FOO_BAR', 'BAZ'):
            with capture() as (out, err):
                _main('-f', '~/${FOO_BAR}_test2.yml', '-p')
            assert out.getvalue() == 'true\n'
        os.remove(TEST2_HOME)

        # Test --dump and task command-line arg passing
        with chdir(os.path.join(DIR, 'foo')):
            with capture() as (out, err):
                _main('--dump')
            assert ('a = abc\nb = abcdef\nc = abcdefghi\n'
                    'pwd = {}\n'.format(DIR)) in out.getvalue()

            _main('print', '5')
            assert read(OUT) == 'abcdefghi 5\n'

            with capture() as (out, err):
                _main()
            assert out.getvalue() == USAGE + '\n'

        # Test task referencing in task definition
        _main('-f', TEST3, 'a')
        assert read(OUT) == 'abc\n'

        _main('-f', TEST3, 'b')
        assert read(OUT) == 'abc\ndef\n'

        # Test --render
        with chdir(DIR):
            with capture() as (out, err):
                _main('--render', 
                      '-i', os.path.join(DIR, 'test1.j2'),
                      '-o', os.path.join(DIR, 'test1.txt'))
            assert out.getvalue() == ''
            assert err.getvalue() == ''

            with open(os.path.join(DIR, 'test1.txt'), 'r') as f:
                txt = f.read()
            assert txt == 'First abc, then abcdef, then abcdefghi.'

            assert_raises(RuntimeError, _main, '--render')
            assert_raises(RuntimeError, _main, '--render', '-i', '/foo/bar/baz')
            assert_raises(RuntimeError, _main, '--render', '-i', 
                          os.path.join(DIR, 'test1.j2'))

        # Test -i and -o path expansion
        assert not (os.path.exists(TEST1_HOME_J2) or 
                    os.path.exists(TEST1_HOME_TXT)), \
            "Run these tests in a Docker container to avoid file clobbering"

        with chdir(DIR):
            shutil.copyfile(os.path.join(DIR, 'test1.j2'), TEST1_HOME_J2)
            shutil.copyfile(os.path.join(DIR, 'test1.txt'), TEST1_HOME_TXT)

            with setitem(os.environ, 'FOO_BAR', 'BAZ'):
                with capture() as (out, err):
                    _main('--render',
                          '-i', '~/${FOO_BAR}_test1.j2',
                          '-o', '~/${FOO_BAR}_test1.txt')

            assert out.getvalue() == ''
            assert err.getvalue() == ''

            with open(TEST1_HOME_TXT, 'r') as f:
                txt = f.read()
            assert txt == 'First abc, then abcdef, then abcdefghi.'

            os.remove(TEST1_HOME_J2)
            os.remove(TEST1_HOME_TXT)

        # Test --pull
        ybase.resolve_url(URL)
        with assign(ybase, 'download', MagicMock()):
            ybase.resolve_url(URL)
            assert ybase.download.call_count == 0

            _main('-f', TEST1)
            assert ybase.download.call_count == 0

            _main('-f', TEST1, '--pull')
            assert ybase.download.call_count == 1

            with tempdir() as d:
                assert_raises(ValidationError, _main, '-f', TEST1, '--cache-dir', d)
            assert ybase.download.call_count == 2

        # Test --validate
        with capture() as (out, err):
            _main('-f', TEST4)
        assert out.getvalue() == ''
        assert err.getvalue() == ''
        assert_raises(UndefinedError, _main, '-f', TEST4, '--validate')

        # Test default_task inheritance
        with capture() as (out, err):
            _main('-f', TEST5, '-p')
        assert out.getvalue() == 'true\n'

        with capture() as (out, err):
            _main('-f', TEST6, '-p')
        assert out.getvalue() == 'false\ntrue\n'

        # Test for and list macros
        with capture() as (out, err):
            _main('-f', TEST7, '-p', 'foo')
        assert out.getvalue() == 'echo x x w 0\n' \
            'echo x x z 1\n' \
            'echo x y w 2\n' \
            'echo x y z 3\n'

        with capture() as (out, err):
            _main('-f', TEST7, '-p', '-s', 'loop_count_macro=count', 'bar')
        assert out.getvalue() == 'echo 1 0\n' \
            'echo 2 1\n' \
            'echo 3 2\n' \
            'echo 4 3\n'

        # Test commands function
        with capture() as (out, err):
            _main('-f', TEST8, '-i', TEST8_IN, '-o', TEST8_OUT, '--render')
        with open(TEST8_OUT, 'r') as f:
            txt = f.read()
            assert txt == '#!/bin/bash\necho foo\necho bar\necho baz\n'

        # Test env function
        with capture() as (out, err):
            _main('-f', TEST8, '-p', 'path')
        assert out.getvalue() == 'echo {}\n'.format(os.environ['PATH'])

        with capture() as (out, err):
            _main('-f', TEST8, '-p', 'bar')
        assert out.getvalue() == 'echo baz\n'

        with capture() as (out, err):
            with setitem(os.environ, 'YATR_BAR', 'foo'):
                _main('-f', TEST8, '-p', 'bar')
        assert out.getvalue() == 'echo foo\n'

        with capture() as (out, err):
            _main('-f', TEST8, '-p', 'baz')
        assert out.getvalue() == 'echo baz_foo foo_bar\n'

        with capture() as (out, err):
            _main('-f', TEST8, 'barfoo')
        assert out.getvalue() == os.path.join(os.environ['PATH'], 'baz_foo') + '\n'

        with capture() as (out, err):
            _main('-f', TEST8, '-p', 'barfoo')
        assert out.getvalue() == ('bar_foo(baz1=baz_foo, path={})\n'
                                  .format(os.environ['PATH']))

        with capture() as (out, err):
            _main('-f', TEST8_2, '-p', 'bar')
        assert out.getvalue() == 'echo baz\n'

        # Test files
        with tempdir() as d:
            _main('-f', TEST9, 'foo', d)
            path = os.path.join(d, 'test1.txt')
            assert os.path.isfile(path)
            with open(path, 'r') as f:
                txt = f.read()
            assert txt == 'foo\n'
            
            with capture() as (out, err):
                _main('-f', TEST9, '-p', 'bar', d)
            assert out.getvalue() == 'cat "{}/test1.txt"\n'.format(d)

        # Test task calling with arguments
        with capture() as (out, err):
            _main('-f', TEST10, '-p', 'x')
        assert out.getvalue() == 'echo 5 3 10\n'

        with capture() as (out, err):
            _main('-f', TEST10, '-p', 'x', '4')
        assert out.getvalue() == 'echo 5 4 4\n'

        with capture() as (out, err):
            _main('-f', TEST10, '-p', 'y', '4')
        assert out.getvalue() == 'echo 7 1 4\n'

        with capture() as (out, err):
            _main('-f', TEST10, '-p', 'z', '4')
        assert out.getvalue() == 'echo 1 6 4\n' \
            'echo 2 6 4\n' \
            'echo 3 6 4\n'

        with capture() as (out, err):
            _main('-f', TEST10, '-p', 'w', '4')
        assert out.getvalue() == 'echo 5 4 4\n' \
            'echo 7 1 4\n' \
            'echo 5 4 4\n'

        # Test anonymous tasks
        with capture() as (out, err):
            _main('-f', TEST10, '-p', 'u')
        assert out.getvalue() == 'echo foo\n' \
            'echo 30 20 10\n'

        # Test _main() calls
        with capture() as (out, err):
            _main('-f', TEST11, 'err')
        assert out.getvalue() == ''
        assert 'does not exist' in err.getvalue()

        with tempfile() as f:
            _main('-f', TEST11, 'bar', 'b', f)
            with open(f, 'r') as f_:
                txt = f_.read().rstrip()
            assert txt == 'b\na\nc\nd\nb'

        with tempfile() as f:
            with capture() as (out, err):
                _main('-f', TEST11, 'baz', 'a', f)
            assert out.getvalue() == ''
            assert 'does not exist' in err.getvalue()

            with capture():
                _main('-f', TEST11, '-s', 'exit_on_error=false', 'baz', 'a', f)
            with open(f, 'r') as f_:
                txt = f_.read().rstrip()
            assert txt == 'a'

        # Test command-line arg validation
        with capture() as (out, err):
            _main('-f', TEST12, '--validate')
        assert out.getvalue() == 'Validation successful\n'

        with capture() as (out, err):
            _main('-f', TEST12, '--validate', 'NONE', 'a')
        assert out.getvalue() == 'Validation successful\n'

        # Test --cache
        with tempdir() as d:
            txt = 'abc123'
            url = 'http://foo.com'
            tmpfile = os.path.join(d, 'temp.txt')
            cached = cached_path(url, d)

            with open(tmpfile, 'w') as f:
                f.write(txt)

            _main('--cache-dir', d, '--cache', '-i', tmpfile, '-o', url)

            with open(cached, 'r') as f:
                out = f.read()
            assert out == txt

            assert_raises(RuntimeError, _main, '--cache')
            assert_raises(RuntimeError, _main, '--cache', '-i', tmpfile)

        # Test --cache-dir path expansion
        assert not os.path.exists(CACHEDIR_HOME), \
            "Run these tests in a Docker container to avoid file clobbering"
        
        os.mkdir(CACHEDIR_HOME)
        d = CACHEDIR_HOME
        txt = 'abc123'
        url = 'http://foo.com'
        tmpfile = os.path.join(d, 'temp.txt')
        cached = cached_path(url, d)
        
        with open(tmpfile, 'w') as f:
            f.write(txt)

        with setitem(os.environ, 'FOO_BAR', 'BAZ'):
            _main('--cache-dir', '~/${FOO_BAR}_cachedir',
                  '--cache', 
                  '-i', '~/${FOO_BAR}_cachedir/temp.txt', 
                  '-o', url)

        with open(cached, 'r') as f:
            out = f.read()
        assert out == txt

        shutil.rmtree(CACHEDIR_HOME)

        # Test dict macros
        with capture() as (out, err):
            _main('-f', TEST13, '--validate')

        with capture() as (out, err):
            _main('-f', TEST13, '-p', 'foo')
            assert out.getvalue() == 'echo def abc\n'

        # Verify example
        with chdir(os.path.join(DIR, 'example')):
            # Test --dump-path
            with capture() as (out, err):
                _main('--dump-path')
            assert out.getvalue() == \
                '{}\n'.format(os.path.join(DIR, 'example/yatrfile.yml'))
            
            # Test -p
            with capture() as (out, err):
                _main('-p', 'bar', 'foo')
            assert out.getvalue() == 'echo bar\necho bar baz foo\n'

            with capture() as (out, err):
                _main('-p', 'cond4')
            assert out.getvalue() == 'ifnot: false\n\techo bar\n'

            # Test --dump
            with capture() as (out, err):
                _main('-f', 'C.yml', '--dump')
            assert 'a = baz\nb = ghi\nc = xyz\n' in out.getvalue()
            
            # Test -m
            with capture() as (out, err):
                _main('-f', 'C.yml', '-m' 'a=zab', '-m', 'd=jkl', '--dump')
            assert 'a = zab\nb = ghi\nc = xyz\nd = jkl\n' in out.getvalue()

            # Cover settings
            _main('-f', 'D.yml', '-s', 'silent=false')

        # Verify render example
        with chdir(os.path.join(DIR, 'example/render')):
            with capture() as (out, err):
                _main('render')
            assert out.getvalue() == ''
            
            with open(DOCKERFILE, 'r') as f:
                txt = f.read()
            assert txt == DFRENDER

            with capture() as (out, err):
                _main('-p', 'build')
            assert out.getvalue() == \
                ('yatr --render -i Dockerfile.j2 -o Dockerfile\n'
                 'docker build -t foo:latest .\n')

        # Validate all test yatrfiles
        with capture() as (out, err):
            _main('-f', TEST1, '--validate')
            _main('-f', TEST2, '--validate')
            _main('-f', TEST3, '--validate')
            # NOTE: TEST4 is intentionally a file that doesn't validate
            _main('-f', TEST5, '--validate')
            _main('-f', TEST6, '--validate')
            _main('-f', TEST7, '--validate')
            _main('-f', TEST8, '--validate')
            _main('-f', TEST8_2, '--validate')
            _main('-f', TEST9, '--validate')
            _main('-f', TEST10, '--validate')
            _main('-f', TEST11, '--validate')
            _main('-f', TEST12, '--validate')
            _main('-f', TEST13, '--validate')
            _main('-f', os.path.join(DIR, 'yatrfile.yml'), '--validate')
            _main('-f', os.path.join(DIR, 'yatrfile1.yml'), '--validate')
            _main('-f', os.path.join(os.path.join(DIR, 'example'), 
                                     'yatrfile.yml'), 
                  '--validate')
            _main('-f', os.path.join(os.path.join(DIR, 'example/render'), 
                                     'yatrfile.yml'), 
                  '--validate')

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



import os
from yatr import Document

DIR = os.path.abspath(os.path.dirname(__file__))
TEST1 = os.path.join(DIR, 'test1.yml')

#-------------------------------------------------------------------------------

def test1():
    test1 = Document.from_path(TEST1)
    assert test1.env.macros['abc'] == 'def'
    assert test1.env.macros['def'] == 'ghi'

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



from yatr import Env
env = Env()

@env.jinja_filter('foo')
def foo(value, **kwargs):
    return '{}_foo'.format(value)

@env.jinja_function('bar')
def bar(value, **kwargs):
    return '{}_bar'.format(value)

@env.task('barfoo', display=('path', 'baz1'))
def bar_foo(env, *args, **kwargs):
    import os
    print(os.path.join(kwargs['path'], kwargs['baz1']))



import yatr.env_decorators as yed
from syn.base_utils import capture
from nose.tools import assert_raises
from yatr import Env

#-------------------------------------------------------------------------------
# EnvDecorator

def test_envdecorator():
    env = Env()
    dec = yed.EnvDecorator(env, 'foo')
    assert_raises(NotImplementedError, dec, None)

#-------------------------------------------------------------------------------
# JinjaFilter

def test_jinjafilter():
    env = Env()
    assert 'foo' not in env.jinja_filters

    @env.jinja_filter('foo')
    def foofilt(value, **kwargs):
        return '{}_foo'.format(value)

    assert foofilt('bar') == 'bar_foo'
    assert env.jinja_filters['foo'] is foofilt

#-------------------------------------------------------------------------------
# JinjaFunction

def test_jinjafunction():
    env = Env()
    assert 'foo' not in env.jinja_functions

    @env.jinja_function('foo')
    def foofunc(value, **kwargs):
        return '{}_foo'.format(value)

    assert foofunc('bar') == 'bar_foo'
    assert env.jinja_functions['foo'] is foofunc

#-------------------------------------------------------------------------------
# Task

def test_task():
    env = Env()
    assert 'foo' not in env.tasks

    @env.task('foo')
    def footask(env, *args, **kwargs):
        pass

    assert footask(env) == 0
    assert env.tasks['foo'].run(env) == [0]

    @env.task('bar')
    def bartask(env, *args, **kwargs):
        raise Exception('bartask exception')

    with capture() as (out, err):
        assert bartask(env) == 1
    assert out.getvalue() == ''
    assert err.getvalue() == 'bartask exception\n'

    with capture() as (out, err):
        assert env.tasks['bar'].run(env) == [1]

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



from jinja2 import UndefinedError
from nose.tools import assert_raises
from syn.base_utils import capture
from yatr import Command, Task, Env, For, ValidationError

#-------------------------------------------------------------------------------
# For

def test_for():
    f = For.from_yaml('x in y')
    assert f.var == 'x'
    assert f.in_ == 'y'
    
    env = Env(env=dict(x='abc', y='def'))
    assert_raises(ValidationError, f.resolve_macros, env)

    env = Env(env=dict(x='abc', y=['d', 'e']))
    assert list(f.resolve_macros(env)) == [['x'], [['d', 'e']]]
    assert list(f.loop(env)) == [Env(env=dict(x='d', y=['d', 'e'])),
                                 Env(env=dict(x='e', y=['d', 'e']))]

    assert_raises(ValidationError, For.from_yaml, '-x in y')
    assert_raises(ValidationError, For.from_yaml, {'var': ['x'], 
                                                   'in': ['x', 'y']})

    f = For.from_yaml({'var': ['x', 'y'], 'in': ['A', 'B']})
    env = Env(env=dict(A=['a', 'b'], B=['c', 'd']))
    assert list(f.resolve_macros(env)) == [['x', 'y'], [['a', 'b'],
                                                        ['c', 'd']]]
    assert list(f.loop(env)) == [Env(env=dict(x='a', y='c', A=['a', 'b'],
                                              B=['c','d'])),
                                 Env(env=dict(x='a', y='d', A=['a', 'b'],
                                              B=['c','d'])),
                                 Env(env=dict(x='b', y='c', A=['a', 'b'],
                                              B=['c','d'])),
                                 Env(env=dict(x='b', y='d', A=['a', 'b'],
                                              B=['c','d']))]

    f = For.from_yaml({'var': ['x', 'y'], 'in': ['A', [1, 2]]})
    env = Env(env=dict(A=['a', 'b']))
    assert list(f.resolve_macros(env)) == [['x', 'y'], [['a', 'b'],
                                                        [1, 2]]]
    assert list(f.loop(env)) == [Env(env=dict(x='a', y=1, A=['a', 'b'])),
                                 Env(env=dict(x='a', y=2, A=['a', 'b'])),
                                 Env(env=dict(x='b', y=1, A=['a', 'b'])),
                                 Env(env=dict(x='b', y=2, A=['a', 'b']))]

    assert_raises(TypeError, For.from_yaml, {})
    assert_raises(ValidationError, For.from_yaml, {'var': ['x', 'y'], 
                                                   'in': ['A', 'B'],
                                                   'bar': 2})

#-------------------------------------------------------------------------------
# Command

def _run_command(c, env):
    with capture() as (out, err):
        c.run(env, preview=True, verbose=True, run_preview=True)
    return out.getvalue().rstrip()

def test_command():
    env = Env()
    c = Command('ls', context='bash')
    assert _run_command(c, env) == 'bash -c "ls"'

    env = Env()
    c = Command('ls')
    assert _run_command(c, env) == 'ls'

    env = Env(macros=dict(a='foo', b='bash'))
    env.resolve_macros()
    c = Command('ls {{a}}', context='{{b}}')
    assert _run_command(c, env) == 'bash -c "ls foo"'
    assert c.command == 'ls {{a}}'

#-------------------------------------------------------------------------------
# Task

def test_task():
    t = Task(commands=[Command('{{a}}'),
                       Command('{{b}}')])

    env = Env()
    assert_raises(UndefinedError, t.run_preview, env)

    env = Env(macros=dict(a='ls', b='pwd'))
    env.resolve_macros()
    assert t.run_preview(env) == 'ls\npwd\n'

    
    t = Task(commands=[Command('ls')])
    assert t == Task.from_yaml('foo', 'ls')

    t = Task(commands=[Command('ls'), Command('pwd')])
    assert t == Task.from_yaml('foo', ['ls', 'pwd'])

    t = Task(commands=[Command('ls', context='bash')])
    assert t == Task.from_yaml('foo', dict(command='ls', context='bash'))

    t = Task(commands=[Command('ls'),
                       Task(commands=[Command('pwd')])])
    assert t == Task.from_yaml('foo', ['ls', dict(task = 'pwd')])
    
    assert_raises(ValidationError, Task.from_yaml, 'foo', [dict(foo='pwd')])

    env = Env()
    t = Task(commands=[Command('1/1', context='python')])
    assert t.run(env) == [0]
    t = Task(commands=[Command('1/0', context='python')])
    with capture():
        assert t.run(env) == [1]

    assert_raises(ValidationError, Task.from_yaml, 'foo', 1)

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



import re

from copy import copy
from functools import partial
from jinja2 import Environment, StrictUndefined
from syn.base import Base, Attr, init_hook
from syn.type import Dict, List, Callable, Set
from syn.five import STR

from .base import resolve, ordered_macros, get_output, DEFAULT_JINJA_FILTERS, \
    DEFAULT_JINJA_FUNCTIONS, fix_functions, variables
from .context import Context, BUILTIN_CONTEXTS
from .task import Task

#-------------------------------------------------------------------------------

CP = 'copy_copy'
UP = 'dict_update'
AUP = 'assign_update'
SUP = 'set_update'

#-------------------------------------------------------------------------------

BUILTIN_PATTERNS = {'_n', '_[0-9]+'} # Conditionally-defined builtin macros
BUILTIN_LISTS = {'ARGS'} # Conditionally-defined builtin list macros

#-------------------------------------------------------------------------------

INITIAL_MACROS = {}

#-------------------------------------------------------------------------------
# DefaultList


class DefaultList(list):
    '''List wrapper that returns a default value for invalid indices.'''
    def __init__(self, lst=[], default=0):
        super(DefaultList, self).__init__(lst)
        self.default_value = default

    def __delitem__(self, idx):
        try:
            super(DefaultList, self).__delitem__(idx)
        except IndexError:
            pass

    def __getitem__(self, idx):
        try:
            return super(DefaultList, self).__getitem__(idx)
        except IndexError:
            return self.default_value


#-------------------------------------------------------------------------------
# Copyable


class Copyable(object):
    def copy(self, **kwargs):
        ret = copy(self)
        for attr in self._groups.copy_copy:
            setattr(ret, attr, copy(getattr(ret, attr)))
        return ret


#-------------------------------------------------------------------------------
# Updateable


class Updateable(object):
    def _update_pre(self, other, **kwargs):
        pass

    def update(self, other, **kwargs):
        self._update_pre(other, **kwargs)

        for attr in self._groups.dict_update:
            getattr(self, attr).update(getattr(other, attr, {}))

        for attr in self._groups.set_update:
            getattr(self, attr).update(getattr(other, attr, set()))

        for attr in self._groups.assign_update:
            value = getattr(other, attr)
            if not value:
                value = getattr(self, attr)
            setattr(self, attr, value)

        self._update_post(other, **kwargs)

    def _update_post(self, other, **kwargs):
        pass


#-------------------------------------------------------------------------------
# Env


class Env(Base, Copyable, Updateable):
    _groups = (UP, AUP)
    _attrs = dict(macros = Attr(Dict((STR, int, float, 
                                      List((STR, int, float, list, dict)),
                                      Dict((STR, int, float, list, dict)))), 
                                init=lambda self: dict(),
                                doc='Macro definitions', groups=(UP, CP)),
                  commandline_macros = Attr(Dict(STR), init=lambda self: dict(),
                                            doc='Macros defined with -m',
                                            groups=(UP, CP)),
                  contexts = Attr(Dict(Context), init=lambda self: dict(),
                                  doc='Execution context definitions', 
                                  groups=(UP, CP)),
                  declares = Attr(set, init=lambda self: set(),
                                  doc='Declared runtime-defined macros',
                                  groups=(SUP, CP)),
                  tasks = Attr(Dict(Task), init=lambda self: dict(),
                               doc='Task definitions', groups=(UP, CP)),
                  secret_values = Attr(Dict(STR), init=lambda self: dict(),
                                       doc='Secret value store', 
                                       groups=(UP, CP)),
                  captures = Attr(Dict(STR), init=lambda self: dict(),
                                  doc='Commands to captures output of', 
                                  groups=(UP, CP)),
                  files = Attr(Dict(STR), init=lambda self: dict(),
                               doc='File name macros', groups=(UP, CP)),
                  settings = Attr(Dict(None), init=lambda self: dict(),
                                  doc='Global settings of various sorts', 
                                  groups=(UP, CP)),
                  jinja_filters = Attr(Dict(Callable), 
                                       init=lambda self: \
                                       dict(DEFAULT_JINJA_FILTERS),
                                       doc='Custom Jinja2 filters', 
                                       groups=(UP, CP)),
                  jinja_functions = Attr(Dict(Callable),
                                         init=lambda self: \
                                         dict(DEFAULT_JINJA_FUNCTIONS),
                                         doc='Custom Jinja2 functions',
                                         groups=(UP, CP)),
                  function_aliases = Attr(Dict(STR), init=lambda self: dict(),
                                          internal=True, groups=(UP, CP),
                                          doc='Jinja function aliases'),
                  env = Attr(Dict((STR, int, float, 
                                   List((STR, int, float, list, dict)),
                                   Dict((STR, int, float, list, dict)))),
                             init=lambda self: dict(),
                             doc='Current name resolution environment', 
                             groups=(UP, CP)),
                  jenv = Attr(Environment, doc='Jinja2 environment', group='eq_exclude',
                              init=lambda self: Environment(undefined=StrictUndefined)),
                  default_task = Attr(STR, '', 'Task to run if no task is '
                                      'specified at the command line', 
                                      group=AUP),
                  default_context = Attr(Context, doc='Execution context to use '
                                         'if none is specified in task definition',
                                         group=AUP))
    _opts = dict(init_validate = True)

    @init_hook
    def _init_populate(self):
        self.macros.update(INITIAL_MACROS)
        self.contexts.update(BUILTIN_CONTEXTS)

        if not hasattr(self, 'default_context'):
            self.default_context = self.contexts['null']

    @init_hook
    def _set_jenv(self, **kwargs):
        filts = dict(self.jinja_filters)
        for name, filt in filts.items():
            filts[name] = partial(filt, env=self)
        self.jenv.filters.update(filts)

        funcs = dict(self.jinja_functions)
        for name, func in funcs.items():
            funcs[name] = partial(func, env=self)
        self.jenv.globals.update(funcs)

    def _update_post(self, other, **kwargs):
        self._set_jenv(**kwargs)

    def capture_value(self, cmd, **kwargs):
        out, code = get_output(cmd)
        # These are intended to be macro values, so newlines and extra
        # white space probably aren't desirable
        return out.strip() 

    def jinja_filter(self, name, *args, **kwargs):
        from .env_decorators import JinjaFilter
        return JinjaFilter(self, name, *args, **kwargs)

    def jinja_function(self, name, *args, **kwargs):
        from .env_decorators import JinjaFunction
        return JinjaFunction(self, name, *args, **kwargs)

    def macro_env(self, **kwargs):
        dct = dict(self.macros)
        dct.update(self.secret_values)
        dct.update(self.files)
        return dct

    def resolve_macros(self, **kwargs):
        env = self.macro_env(**kwargs)
        macros = dict(self.macros)
        macros.update(self.captures)
        macros.update(self.commandline_macros)

        # TODO: better error message if there is a cycle
        potential_problems = set(env) & set(self.jinja_functions)
        for name, template in ordered_macros(macros, jenv=self.jenv,
                                             funcs=self.jinja_functions):
            if name in self.macros or name in self.commandline_macros:
                fixed = fix_functions(template, potential_problems, self)
                #env[name] = resolve(fixed, env, jenv=self.jenv)
                env[name] = self.resolve(fixed, env=env, **kwargs)

            if name in self.captures:
                cmd = resolve(template, env, jenv=self.jenv)
                env[name] = self.capture_value(cmd, **kwargs)

        self.env = env

    def resolve(self, template, **kwargs):
        env = dict(kwargs.get('env', self.env))

        if kwargs.get('from_validate', False):
            names = variables(template, jenv=self.jenv)

            patterns = set(BUILTIN_PATTERNS)
            patterns.update(self.declares)
            patterns = [re.compile(p) for p in patterns]

            # If the builtin/declared macro isn't defined, set it to ''
            # so that it can validate
            for name in names:
                if name not in env:
                    for p in patterns:
                        if re.match(p, name):
                            env[name] = ''
                            break

                if name in BUILTIN_LISTS:
                    lst = list(env.get(name, []))
                    env[name] = DefaultList(lst)

        return resolve(template, env, jenv=self.jenv)

    def task(self, name, *args, **kwargs):
        from .env_decorators import Task
        return Task(self, name, *args, **kwargs)

    def validate(self):
        super(Env, self).validate()
        
        for name, ctx in self.contexts.items():
            ctx.validate()

        for name, task in self.tasks.items():
            task.validate()


#-------------------------------------------------------------------------------
# __all__

__all__ = ('Env',)

#-------------------------------------------------------------------------------
